#
# Copyright (C) 2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import sys
import pathlib
import datetime


GITHUB_USER_REPO = 'UAVCAN', 'pydsdl'

DESCRIPTION = 'A UAVCAN DSDL compiler front-end in Python'

DOC_ROOT = pathlib.Path(__file__).absolute().parent
REPOSITORY_ROOT = DOC_ROOT.parent

sys.path.insert(0, str(REPOSITORY_ROOT))
import pydsdl
assert 'site-packages' not in pydsdl.__file__, 'Wrong import source'

PACKAGE_ROOT = pathlib.Path(pydsdl.__file__).absolute().parent

EXTERNAL_LINKS = {
    'UAVCAN homepage': 'https://uavcan.org/',
    'Support forum':   'https://forum.uavcan.org/',
}

project = 'PyDSDL'
# noinspection PyShadowingBuiltins
copyright = str(datetime.datetime.now().year) + ', UAVCAN Development Team'
author = 'UAVCAN Development Team'

version = '.'.join(map(str, pydsdl.__version_info__))
release = pydsdl.__version__  # The full version, including alpha/beta/rc tags

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.coverage',
    'sphinx.ext.todo',
    'sphinx.ext.intersphinx',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.graphviz',
    'sphinx_computron',
    'ref_fixer_hack',
]
sys.path.append(str(DOC_ROOT))  # This is for the hack to be importable

templates_path = []

exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

source_suffix = ['.rst']

master_doc = 'index'

autoclass_content = 'class'
autodoc_member_order = 'bysource'
autodoc_inherit_docstrings = False
autodoc_default_options = {
    'members':          True,
    'undoc-members':    True,
    'special-members':  True,
    'imported-members': True,
    'show-inheritance': True,
    'member-order':     'bysource',
    'exclude-members':
        '__weakref__, __module__, __dict__, __dataclass_fields__, __dataclass_params__, __annotations__, '
        '__abstractmethods__, __orig_bases__, __parameters__, __post_init__, __getnewargs__',
}

todo_include_todos = True

graphviz_output_format = 'svg'

inheritance_graph_attrs = {
    'rankdir':  'LR',
    'bgcolor':  '"transparent"',
}
inheritance_node_attrs = {
    'color':     '"#000000"',
    'fontcolor': '"#000000"',
}
inheritance_edge_attrs = {
    'color': inheritance_node_attrs['color'],
}

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
}

pygments_style = 'friendly'

html_favicon = '_static/favicon.ico'
html_theme = 'sphinx_rtd_theme'
html_theme_options = {
    'display_version':            True,
    'prev_next_buttons_location': 'bottom',
    'style_external_links':       True,
    'navigation_depth':           -1,
}
html_context = {
}
html_static_path = ['_static']
html_css_files = [
    'custom.css',
]



"""
========================================  THIS IS A DIRTY HACK  ========================================

I've constructed this Sphinx extension as a quick and dirty "solution" to the problem of broken cross-linking.

The problem is that Autodoc fails to realize that an entity, say, pyuavcan.transport._session.InputSession is
exposed to the user as pyuavcan.transport.InputSession, and that the original name is not a part of the API
and it shouldn't even be mentioned in the documentation at all. I've described this problem in this Sphinx issue
at https://github.com/sphinx-doc/sphinx/issues/6574. Since the original name is not exported, Autodoc can't find
it in the output and generates no link at all, requiring the user to search manually instead of just clicking on
stuff.

The hack is known to occasionally misbehave and produce incorrect links at the output, but hey, it's a hack.
Someone should just fix Autodoc instead of relying on this long-term. Please.
"""

import re
import os
import typing

import sphinx.application
import sphinx.environment
import sphinx.util.nodes
import docutils.nodes


_ACCEPTANCE_PATTERN = r'.*([a-zA-Z][a-zA-Z0-9_]*\.)+_[a-zA-Z0-9_]*\..+'
_REFTYPES = 'class', 'meth', 'func'

_replacements_made: typing.List[typing.Tuple[str, str]] = []


def missing_reference(app:      sphinx.application.Sphinx,
                      _env:     sphinx.environment.BuildEnvironment,
                      node:     docutils.nodes.Element,
                      contnode: docutils.nodes.Node) -> typing.Optional[docutils.nodes.Node]:
    old_reftarget = node['reftarget']
    if node['reftype'] in _REFTYPES and re.match(_ACCEPTANCE_PATTERN, old_reftarget):
        new_reftarget = re.sub(r'\._[a-zA-Z0-9_]*', '', old_reftarget)
        if new_reftarget != old_reftarget:
            _replacements_made.append((old_reftarget, new_reftarget))
            attrs = contnode.attributes if isinstance(contnode, docutils.nodes.Element) else {}
            try:
                old_refdoc = node['refdoc']
            except KeyError:
                return None
            new_refdoc = old_refdoc.rsplit(os.path.sep, 1)[0] + os.path.sep + new_reftarget.rsplit('.', 1)[0]
            return sphinx.util.nodes.make_refnode(app.builder,
                                                  old_refdoc,
                                                  new_refdoc,
                                                  node.get('refid', new_reftarget),
                                                  docutils.nodes.literal(new_reftarget, new_reftarget, **attrs),
                                                  new_reftarget)
    return None


def doctree_resolved(_app:     sphinx.application.Sphinx,
                     doctree:  docutils.nodes.document,
                     _docname: str) -> None:
    def predicate(n: docutils.nodes.Node) -> bool:
        if isinstance(n, docutils.nodes.FixedTextElement):
            is_text_primitive = len(n.children) == 1 and isinstance(n.children[0], docutils.nodes.Text)
            if is_text_primitive:
                return is_text_primitive and re.match(_ACCEPTANCE_PATTERN, n.children[0].astext())
        return False

    def substitute_once(text: str) -> str:
        out = re.sub(r'\._[a-zA-Z0-9_]*', '', text)
        _replacements_made.append((text, out))
        return out

    # The objective here is to replace all references to hidden objects with their exported aliases.
    # For example: pyuavcan.presentation._typed_session._publisher.Publisher --> pyuavcan.presentation.Publisher
    for node in doctree.traverse(predicate):
        assert isinstance(node, docutils.nodes.FixedTextElement)
        node.children = [
            docutils.nodes.Text(substitute_once(node.children[0].astext()))
        ]


def setup(app: sphinx.application.Sphinx):
    app.connect('missing-reference', missing_reference)
    app.connect('doctree-resolved', doctree_resolved)
    return {
        'parallel_read_safe': True,
    }



"""
This script is used for generating documentation for the specified class and all its descendants.
It accepts one argument: the name of the class, and prints the ReST documentation markup.
The script is intended to be invoked from the ReST documents using https://github.com/pavel-kirienko/sphinx-computron.
"""

import sys
import itertools
import pydsdl


def children(ty):
    for t in ty.__subclasses__():
        yield t
        yield from children(t)


T = getattr(pydsdl, sys.argv[1])
for t in itertools.chain([T], children(pydsdl.Any)):
    print('.. autoclass::', '.'.join([t.__module__, t.__name__]))
    print('   :members:')
    print('   :undoc-members:')
    print('   :special-members:')
    print('   :no-inherited-members:')
    print('   :show-inheritance:')
    print()



"""
This script is used for generating inheritance diagrams in the docs.
It accepts one argument: the name of the class, and prints the ReST markup into stdout that produces the diagram.
The script is intended to be invoked from the ReST documents using https://github.com/pavel-kirienko/sphinx-computron.
"""

import sys
import pydsdl


def children(ty):
    for t in ty.__subclasses__():
        yield t
        yield from children(t)


T = getattr(pydsdl, sys.argv[1])
print('.. inheritance-diagram::', ' '.join('.'.join([t.__module__, t.__name__]) for t in children(T)))
print('   :parts: 1')
print('   :top-classes:', '.'.join([T.__module__, T.__name__]))
